/**
 * @file SeedService_service.cc
 * @brief Assists in the distribution of guaranteed unique seeds to all engines within a job.
 * @author Rob Kutschke (kutschke@fnal.gov)
 * @see SeedService.hh SeedMaster.hh
 */

// SeedService header
#include "artextensions/SeedService/SeedService.hh"

// Art include files
#include "art/Utilities/Exception.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "art/Persistency/Provenance/ModuleDescription.h"
#include "art/Persistency/Provenance/EventID.h"

// Supporting library include files
#include "messagefacility/MessageLogger/MessageLogger.h"

// C++ include files
#include <iostream>
#include <iomanip>

namespace artext {

  //----------------------------------------------------------------------------
  SeedService::SeedService
    (fhicl::ParameterSet const& paramSet, art::ActivityRegistry& iRegistry)
    : seeds(paramSet)
    , state()
    , verbosity(paramSet.get<int>("verbosity", 0))
    , bPrintEndOfJobSummary(paramSet.get<bool>("endOfJobSummary",false))
  {
    // Register callbacks.
    iRegistry.sPreModuleConstruction.watch  (this, &SeedService::preModuleConstruction  );
    iRegistry.sPostModuleConstruction.watch (this, &SeedService::postModuleConstruction );
    iRegistry.sPreModuleBeginRun.watch      (this, &SeedService::preModuleBeginRun      );
    iRegistry.sPostModuleBeginRun.watch     (this, &SeedService::postModuleBeginRun     );
    iRegistry.sPostEndJob.watch             (this, &SeedService::postEndJob             );
  } // SeedService::SeedService()
  
  
  //----------------------------------------------------------------------------
  SeedService::seed_t SeedService::getSeed() {
    return getSeed(EngineId(state.currentModuleLabel));
  } // SeedService::getSeed()
  
  
  //----------------------------------------------------------------------------
  SeedService::seed_t SeedService::getSeed(std::string instanceName) {
    return getSeed(EngineId(state.currentModuleLabel, instanceName));
  } // SeedService::getSeed(string)
  
  
  //----------------------------------------------------------------------------
  void SeedService::print() const { seeds.print(); }
  
  
  //----------------------------------------------------------------------------
  SeedService::seed_t SeedService::getSeed(EngineId const& id) {
    
    // Are we being called from the right place?
    ensureValidState();
    
    // Ask the seed master about seeds...
    return seeds.getSeed(id);
  } // SeedService::getSeed(EngineID)
  
  
  //----------------------------------------------------------------------------
  void SeedService::ensureValidState() const {
    // getSeed may only be called from a c'tor or from a beginRun method.
    // In all other cases, throw.
    if ( state.state == SeedServiceHelper::ArtState::unDefined ) {
      throw art::Exception(art::errors::LogicError)
        << "SeedService: not in a module constructor or beginRun method. May not call getSeed.";
    }
  } // SeedService::ensureValidState()
  
  
  //----------------------------------------------------------------------------
  // Callbacks called by art.  Used to maintain information about state.
  void SeedService::preModuleConstruction(art::ModuleDescription const& md)  {
    state.set(SeedServiceHelper::ArtState::inConstructor, md.moduleLabel());
  } // SeedService::preModuleConstruction()
  
  void SeedService::postModuleConstruction(art::ModuleDescription const&) {
    state.clear();
  } // SeedService::postModuleConstruction()
  
  void SeedService::preModuleBeginRun(art::ModuleDescription const& md) {
    state.set(SeedServiceHelper::ArtState::inBeginRun, md.moduleLabel());
  } // SeedService::preModuleBeginRun()
  
  void SeedService::postModuleBeginRun(art::ModuleDescription const&) {
    state.clear();
  } // SeedService::postModuleBeginRun()
  
  
  void SeedService::postEndJob() {
    if ((verbosity > 0) || bPrintEndOfJobSummary)
      print(); // framework logger decides whether and where it shows up
  } // SeedService::postEndJob()
  
  //----------------------------------------------------------------------------

} // end namespace artext

DEFINE_ART_SERVICE(artext::SeedService)
